from __future__ import annotations

import logging
from dataclasses import fields
from enum import Enum
from typing import Any, Dict, List, NamedTuple, Optional, Tuple

from blinker import Signal

import ieee_2030_5.hrefs as hrefs
import ieee_2030_5.models as m
from ieee_2030_5.adapters import (Adapter, AdapterListProtocol, BaseAdapter,
                                  ready_signal)
from ieee_2030_5.adapters.timeadapter import TimeAdapter
from ieee_2030_5.config import InvalidConfigFile
from ieee_2030_5.data.indexer import add_href, get_href_filtered
from ieee_2030_5.models.sep import DERProgram
from ieee_2030_5.types_ import StrPath

_log = logging.getLogger(__name__)

__all__: List[str] = [
    "DERControlAdapter",
    "DERProgramAdapter",
    "DERCurveAdapter",
    "DERAdapter"
]

class CreateStatus(Enum):
    Error = "Error"
    Created = "Created"
    Updated = "Updated"

class CreateResponse(NamedTuple):
    data: Any
    href: str
    index: int = -1
    status: str = CreateStatus.Created.value
    
    @property
    def statusint(self) -> int:
        if self.status == CreateStatus.Created.value:
            return 201
        elif self.status == CreateStatus.Updated.value:
            return 204
        elif self.status == CreateStatus.Error:
            return 400
        

DERCurveAdapter = Adapter[m.DERCurve]("curve", generic_type=m.DERCurve)

def initialize_der_curve_adapter(sender):
    config = BaseAdapter.server_config()
    for index, curve_cfg in enumerate(config.curves):
        der_curve = m.DERCurve(**curve_cfg.__dict__)
        der_curve.href = hrefs.curve_href(index)
        DERCurveAdapter.add(der_curve)
    
    ready_signal.send(DERCurveAdapter)
    
ready_signal.connect(initialize_der_curve_adapter, BaseAdapter)


DERControlAdapter = Adapter[m.DERControl]("derc", generic_type=m.DERControl)

def initialize_der_control_adapter(sender):
    config = BaseAdapter.server_config()
    for index, ctl in enumerate(config.controls):
        
        # Create a new DERControl and DERControlBase and initialize as much as possible
        base_control: m.DERControlBase = BaseAdapter.build_instance(m.DERControlBase, ctl.base)

        control: m.DERControl = BaseAdapter.build_instance(m.DERControl, ctl.__dict__)
        control.href = hrefs.get_derc_href(index=index)
        #control.mRID = f"MYCONTROL{index}"
        control.DERControlBase = base_control
        
        DERControlAdapter.add(control)
    
    ready_signal.send(DERControlAdapter)
ready_signal.connect(initialize_der_control_adapter, DERCurveAdapter)


DERProgramAdapter = Adapter[DERProgram](hrefs.der_program_href(), generic_type=m.DERProgram)

def time_updated(timestamp):
    
    for derp_index, derp in enumerate(DERProgramAdapter.fetch_all()):
        try:            
            controls = DERProgramAdapter.fetch_children(derp, hrefs.DERC)
        except KeyError:
            # If controls aren't defined in the program then no need to continue
            continue
        
        try:
            current_active = DERProgramAdapter.fetch_children(derp, hrefs.DER_CONTROL_ACTIVE)
        except KeyError:
            current_active = []
            
        for ctrl_index, ctrl in enumerate(controls):
            if not ctrl.EventStatus:
                if ctrl.interval is not None:
                    _log.debug(f"Setting up event for ctrl {ctrl_index} current_time: {timestamp} start_time: {ctrl.interval.start}")
                    if timestamp > ctrl.interval.start and timestamp < ctrl.interval.start + ctrl.interval.duration:
                        ctrl.EventStatus = m.EventStatus(currentStatus=1, dateTime=timestamp, potentiallySuperseded=False, reason="Active") 
                    else:
                        ctrl.EventStatus = m.EventStatus(currentStatus=0, dateTime=timestamp, potentiallySuperseded=False, reason="Scheduled") 
                    _log.debug(f"ctrl.EventStatus is {ctrl.EventStatus}")
                else:
                    ctrl.EventStatus = m.EventStatus(currentStatus=0, dateTime=timestamp, potentiallySuperseded=False, reason="Scheduled") 
                
            if ctrl.EventStatus:
                if ctrl.interval:
                    # Active control
                    if ctrl.interval.start < timestamp and timestamp < ctrl.interval.start + ctrl.interval.duration:
                        if ctrl.EventStatus.currentStatus == 0:
                            _log.debug(f"Activating control {ctrl_index}")
                            ctrl.EventStatus.currentStatus = 1 # Active
                            ctrl.EventStatus.dateTime = timestamp
                            ctrl.EventStatus.reason = f"Control event active {ctrl.mRID}"

                        if ctrl.mRID not in [x.mRID for x in current_active]:
                            DERProgramAdapter.add_replace_child(derp, hrefs.DER_CONTROL_ACTIVE, ctrl)
                            
                    elif timestamp > ctrl.interval.start + ctrl.interval.duration:
                        if ctrl.EventStatus.currentStatus == 1:
                            _log.debug(f"Deactivating control {ctrl_index}")
                            
                            ctrl.EventStatus.currentStatus = -1 # for me this means complete
                            DERProgramAdapter.remove_child(derp, hrefs.DER_CONTROL_ACTIVE, ctrl)
                

def initialize_der_program_adapter(sender):
    
    config = BaseAdapter.server_config()
    
    controls: List[m.DERControl] = DERControlAdapter.fetch_all()
    curves: List[m.DERCurve] = DERCurveAdapter.fetch_all()    
   
    # Initialize "global" m.DERPrograms href lists, including all the different links to
    # locations for active, default, curve and control lists.
    for index, program_cfg in enumerate(config.programs):
        # The configuration contains a mapping to control lists so when
        # building the DERProgram object we need to remove it from the paramters before
        # initialization.
        params = program_cfg.__dict__.copy()
        del params['default_control']
        del params['controls']
        del params['curves']
        
        program = m.DERProgram(**params)
        program.description = program_cfg.description
        program.primacy = program_cfg.primacy
        # program.version = program_cfg.version

        # TODO Fix this!
        # program.mRID =
        # mrid = program_cfg.get('mrid')
        # if mrid is None or len(mrid.trim()) == 0:
        #     program.mRID = f"program_mrid_{index}"
        # program_cfg['mrid'] = program.mRID

        program.href = hrefs.get_program_href(index)
        program.DefaultDERControlLink = m.DefaultDERControlLink(href=hrefs.der_program_href(index, hrefs.DERProgramSubType.DefaultDERControlLink))
        program.ActiveDERControlListLink = m.ActiveDERControlListLink(href=hrefs.der_program_href(index, hrefs.DERProgramSubType.ActiveDERControlListLink))
        program.DERControlListLink = m.DERControlListLink(href=hrefs.der_program_href(index, hrefs.DERProgramSubType.DERControlListLink))
        program.DERCurveListLink = m.DERCurveListLink(href=hrefs.der_program_href(index, hrefs.DERProgramSubType.DERCurveListLink))
        DERProgramAdapter.add(program)
        
        default_ctl = None
        
        for ctl_cfg in program_cfg.controls:
            for ctl in controls:
                if isinstance(ctl_cfg, str):
                    description = ctl_cfg
                else:
                    description = ctl_cfg["description"]
                if ctl.description == description:
                    DERProgramAdapter.add_replace_child(program, hrefs.DERC, ctl)

        for ctl in controls:
            if ctl.description == program_cfg.default_control:
                DERProgramAdapter.add_replace_child(program, hrefs.DDERC, ctl)
                default_ctl = ctl
        
        if default_ctl is None:
            raise InvalidConfigFile(f"Section program: {program_cfg.description} default control {program_cfg.default_control} not found!")
            
        for curve_cfg in program_cfg.curves:
            for curve in curves:
                if isinstance(curve_cfg, str):
                    description = curve_cfg
                else:
                    description = curve_cfg["description"]
                    
                if curve.description == description:
                    DERProgramAdapter.add_replace_child(program, "dc", curve)
        
        ready_signal.send(DERProgramAdapter)
        TimeAdapter.tick.connect(time_updated)
        # else:
        #     default_ctl: m.DefaultDERControl = BaseAdapter.build_instance(
        #         m.DefaultDERControl, der_cfg.__dict__)
        #     default_ctl.href = hrefs.get_derc_default_href(index)
        #     #default_ctl.mRID = der_ctl.mRID + " default"
        #     default_ctl.setESDelay = 20
        #     default_ctl.DERControlBase = der_ctl.DERControlBase

        #     program.DefaultDERControlLink = m.DefaultDERControlLink(href=default_ctl.href)
        #     self._der_programs.append(program)
        #     program.ActiveDERControlListLink = m.ActiveDERControlListLink(href=hrefs.der_program_href(index, hrefs.DERProgramSubType.ActiveDERControlListLink))
        #     program.DERControlListLink = m.DERControlListLink(href=hrefs.der_program_href(index, hrefs.DERProgramSubType.DERControlListLink.value))
        #     program.DERCurveListLink = m.DERCurveListLink(href=hrefs.der_program_href(index, hrefs.DERProgramSubType.DERCurveListLink.value))
            

        # der_control_list = m.DERControlList(href=hrefs.get_program_href(index, hrefs.DERC))

        # for ctl_description in program_cfg.controls:
        #     try:
        #         derc = next(filter(lambda d: d.description == ctl_description, der_controls))
        #     except StopIteration:
        #         raise InvalidConfigFile(
        #             f"Section program: {program_cfg.description} control {ctl_description} not found!"
        #         )
        #     else:
        #         der_control_list.DERControl.append(derc)

        # add_href(der_control_list.href, der_control_list)

        # der_curve_list = m.DERCurveList(href=hrefs.get_program_href(index, hrefs.CURVE))

        # for curve_description in program_cfg.curves:
        #     try:
        #         der_curve = next(
        #             filter(lambda d: d.description == curve_description, der_curves))
        #     except StopIteration:
        #         raise InvalidConfigFile(
        #             f"Section program: {program_cfg.description} curve {curve_description} not found!"
        #         )
        #     else:
        #         der_curve_list.DERCurve.append(der_curve)

        # der_curve_list.all = len(der_curve_list.DER)
        #ready_signal.send(DERProgramAdapter)
    

ready_signal.connect(initialize_der_program_adapter, DERControlAdapter)
# DERProgramAdapter = _DERProgramAdapter()
# ready_signal.connect(DERProgramAdapter.__initialize__, DERControlAdapter)


# class _DERAdapter(BaseAdapter, AdapterListProtocol):
#     def __init__(self) -> None:
#         super().__init__()
        
#         self._edev_ders: Dict[int, List[m.DER]] = {}
#         self._der_capabilities: Dict[int, List[m.DERCapability]] = {}
#         self._der_settings: Dict[int, List[m.DERSettings]] = {}
#         self._der_status: Dict[int, List[m.DERStatus]] = {}
#         self._der_availabilites: Dict[int, List[m.DERAvailability]] = {}
#         self._der_current_program: Dict[int, List[m.DERProgram]] = {}
#         self._edev_der_settings: Dict[int, List[m.DERSettings]] = {}
        
#     def __initialize__(self, sender):
#         # TODO: Load ders
#         cfg = BaseAdapter.server_config()
        
#     def create(self, edev_index: int, der_index: int,  modesSupported: str, deviceType: int) -> m.DER:
        
#         der = m.DER(href=hrefs.edev_der_href(edev_index=edev_index, der_index=der_index))
#         der.DERCapabilityLink = m.DERCapabilityLink(hrefs.der_sub_href(edev_index=edev_index, index=der_index, subtype=hrefs.DERSubType.Capability))
#         der.DERSettingsLink = m.DERSettingsLink(hrefs.der_sub_href(edev_index=edev_index, index=der_index, subtype=hrefs.DERSubType.Settings))
#         der.DERStatusLink = m.DERStatusLink(hrefs.der_sub_href(edev_index=edev_index, index=der_index, subtype=hrefs.DERSubType.Status))
#         der.DERAvailabilityLink = m.DERAvailabilityLink(hrefs.der_sub_href(edev_index=edev_index, index=der_index, subtype=hrefs.DERSubType.Availability))
#         # der.CurrentDERProgramLink = m.CurrentDERProgramLink(hrefs.der_sub_href(edev_index=edev_index, index=der_index, subtype=hrefs.DERSubType.CurrentProgram))
#         if not self._edev_ders.get(edev_index):
#             self._edev_ders[edev_index] = []
            
#         self._edev_ders[edev_index].append(der)
        
#         if not self._der_capabilities.get(edev_index):
#             self._der_capabilities[edev_index] = []
#         self._der_capabilities[edev_index].append(m.DERCapability(hrefs.der_sub_href(edev_index=edev_index, index=der_index, subtype=hrefs.DERSubType.Capability),
#                                                              modesSupported=modesSupported,
#                                                              type=deviceType))
        
#         if not self._der_settings.get(edev_index):
#             self._der_settings[edev_index] = []
            
#         self._der_settings[edev_index].append(m.DERSettings(href=hrefs.der_sub_href(edev_index=edev_index, index=der_index, subtype=hrefs.DERSubType.Settings)))
        
#         if not self._der_status.get(edev_index):
#             self._der_status[edev_index] = []
        
#         self._der_status[edev_index].append(m.DERStatus(href=hrefs.der_sub_href(edev_index=edev_index, index=der_index, subtype=hrefs.DERSubType.Status)))
        
#         if not self._der_availabilites.get(edev_index):
#             self._der_availabilites[edev_index] = []
        
#         self._der_availabilites[edev_index].append(m.DERAvailability(href=hrefs.der_sub_href(edev_index=edev_index, index=der_index, subtype=hrefs.DERSubType.Availability)))
        
#         if not self._der_current_program.get(edev_index):
#             self._der_current_program[edev_index] = []
        
#         self._der_current_program[edev_index].append(m.DERProgram(href=hrefs.der_sub_href(edev_index=edev_index, index=der_index, subtype=hrefs.DERSubType.CurrentProgram)))
        
#         return der
    
#     def fetch_edev_all(self, edev_index: int) -> List:
#         return self._edev_ders.get(edev_index, [])
    
#     def fetch_list(self, edev_index: int, start: int = 0, after: int = 0, limit: int = 0) -> m.DERList:
        
#         try:
#             der_list = m.DERList(href=hrefs.der_sub_href(edev_index), 
#                                 all=len(self._edev_ders[edev_index]), results=len(self._edev_ders[edev_index]), DER=self._edev_ders[edev_index])
#         except KeyError:
#             der_list = m.DERList(href=hrefs.der_sub_href(edev_index), 
#                                  all=0, results=0, DER=[])
#         return der_list
    
#     def fetch_at(self, edev_index: int, der_index: int) -> m.DER:
#         return self._edev_ders[edev_index][der_index]
    
#     def fetch_settings_at(self, edev_index: int, der_index: int) -> m.DERSettings:
#         return self._edev_der_settings[edev_index][der_index]
    
#     def store_settings_at(self, edev_index: int, der_index: int, settings: m.DERSettings):
#         self._edev_der_settings[edev_index][der_index] = settings
    
#     def fetch_status_at(self, edev_index: int, der_index: int) -> m.DERStatus:
#         return self._der_status[edev_index][der_index]
    
#     def store_status_at(self, edev_index: int, der_index: int, status: m.DERStatus):
#         self._edev_der_status[edev_index][der_index] = status
    
#     def fetch_capability_at(self, edev_index: int, der_index: int) -> m.DERCapability:
#         return self._der_capabilities[edev_index][der_index]
    
#     def store_capability_at(self, edev_index: int, der_index: int, capability: m.DERCapability):
#         self._der_capabilities[edev_index][der_index] = capability
    
#     def fetch_availibility_at(self, edev_index: int, der_index: int) -> m.DERAvailability:
#         return self._der_availabilites[edev_index][der_index]
    
#     def store_availibility_at(self, edev_index: int, der_index: int, availability: m.DERAvailability):
#         self._der_availabilites[edev_index][der_index] = availability
    
#     def fetch_current_program_at(self, edev_index: int, der_index: int) -> m.DERProgram:
#         return self._der_current_program[edev_index][der_index]
    
#     def store_current_program_at(self, edev_index: int, der_index: int, current_program: m.DERProgram):
#         self._der_current_program[edev_index][der_index] = current_program
    
#     def store(self, parsed_edev: hrefs.EdevHref, data) -> int:
#         if parsed_edev.der_sub == hrefs.DERSubType.Availability.value:
#             self._der_availabilites[parsed_edev.edev_index][parsed_edev.der_index] = data
#         elif parsed_edev.der_sub == hrefs.DERSubType.Capability.value:
#             self._der_capabilities[parsed_edev.edev_index][parsed_edev.der_index] = data
#         elif parsed_edev.der_sub == hrefs.DERSubType.Status.value:
#             self._der_status[parsed_edev.edev_index][parsed_edev.der_index] = data
#         elif parsed_edev.der_sub == hrefs.DERSubType.Settings.value:
#             self._der_settings[parsed_edev.edev_index][parsed_edev.der_index] = data
#         else:
#             raise ValueError(data)
        
#         return 200
    
#     def get_list(self, edev_index: int):
#         return self._edev_ders[edev_index]
    
    
# DERAdapter = _DERAdapter()
# ready_signal.connect(DERAdapter.__initialize__, BaseAdapter)



if __name__ == '__main__':
    from pathlib import Path

    import yaml

    from ieee_2030_5.__main__ import get_tls_repository
    from ieee_2030_5.config import ServerConfiguration
    
    cfg_pth = Path("/home/os2004/repos/gridappsd-2030_5/config.yml")
    cfg_dict = yaml.safe_load(cfg_pth.read_text())

    config = ServerConfiguration(**cfg_dict)

    tls_repo = get_tls_repository(config, False)

    BaseAdapter.initialize(config, tls_repo)    

    print(DERCurveAdapter.fetch_edev_all())
    print(DERCurveAdapter.fetch_list())


import logging
from functools import lru_cache
from typing import List

import ieee_2030_5.hrefs as hrefs
import ieee_2030_5.models as m
from ieee_2030_5.adapters import BaseAdapter
from ieee_2030_5.types_ import Lfdi

_log = logging.getLogger(__name__)


__all__: List[str] = [
    "DeviceCapabilityAdapter"
]

class _DeviceCapabilityAdapter:
    
    def __init__(self) -> None:
        pass
    
    def __after_base_init__(self, sender):
        pass

    @classmethod
    @lru_cache
    def get_default_dcap(cls) -> m.DeviceCapability:
        dcap = m.DeviceCapability(href=hrefs.get_dcap_href(),
                                  pollRate=BaseAdapter.server_config().device_capability_poll_rate)
        dcap.ResponseSetListLink = m.ResponseSetListLink(href=hrefs.get_response_set_href(), all=0)
        dcap.TimeLink = m.TimeLink(href=hrefs.get_time_href())
        dcap.EndDeviceListLink = m.EndDeviceListLink(href=hrefs.get_enddevice_href(hrefs.NO_INDEX),
                                                     all=0)
        dcap.MirrorUsagePointListLink = m.MirrorUsagePointListLink(
            href=hrefs.mirror_usage_point_href())
        dcap.UsagePointListLink = m.UsagePointListLink(href=hrefs.usage_point_href())
        return dcap

    @staticmethod
    def get_by_lfdi(lfdi: Lfdi) -> m.DeviceCapability:
        dc = DeviceCapabilityAdapter.get_default_dcap()
        config = BaseAdapter.get_config_from_lfdi(lfdi)
        if config:
            dc.EndDeviceListLink.all = 1
        return dc

DeviceCapabilityAdapter = _DeviceCapabilityAdapter()
BaseAdapter.after_initialized.connect(DeviceCapabilityAdapter.__after_base_init__)


import logging
import uuid
from ast import Dict
from datetime import datetime
from typing import List

import ieee_2030_5.hrefs as hrefs
import ieee_2030_5.models as m
from ieee_2030_5.adapters import (Adapter, AdapterListProtocol, BaseAdapter,
                                  ready_signal)
from ieee_2030_5.adapters.der import DERProgramAdapter
from ieee_2030_5.adapters.fsa import FSAAdapter
from ieee_2030_5.adapters.timeadapter import TimeAdapter
from ieee_2030_5.data.indexer import add_href
from ieee_2030_5.models.enums import DeviceCategoryType
from ieee_2030_5.types_ import Lfdi
from ieee_2030_5.utils import uuid_2030_5

_log = logging.getLogger(__file__)


EndDeviceAdapter = Adapter[m.EndDevice](hrefs.get_enddevice_href(), generic_type=m.EndDevice)
def initialize_end_device_adapter(sender):
    """ Intializes the following based upon the device configuration and the tlsrepository.
        
    Each EndDevice will have the following sub-components initialized:
    - PowerStatus - PowerStatusLink
    - DeviceStatus - DeviceStatusLink
    - Registration - RegistrationLink
    - MessagingProgramList - MessagingProgramListLink
    - Log
    Either FSA or DemandResponseProgram
    - DemandResponseProgram - DemandResponseProgramListLink
    
    
    As well as the following properties
    - changedTime - Current time of initialization
    - sFDI - The short form of the certificate for the system.
    """

    # assert EndDeviceAdapter.__tls_repository__ is not None
    # EndDeviceAdapter.initialize_from_storage()
    # programs = DERProgramAdapter.get_all()
    # stored_devices = EndDeviceAdapter.get_all()
    programs = DERProgramAdapter.fetch_all()

    for dev in BaseAdapter.device_configs():
        ts = int(round(datetime.utcnow().timestamp()))
        
        edev = m.EndDevice()
        edev.lFDI = BaseAdapter.__tls_repository__.lfdi(dev.id)
        edev.sFDI = BaseAdapter.__tls_repository__.sfdi(dev.id)
        # TODO handle enum eval in a better way.
        edev.deviceCategory = eval(f"DeviceCategoryType.{dev.deviceCategory}")
        edev.enabled = dev.enabled
        edev.changedTime = ts

        # TODO remove subscribable
        edev.subscribable = None
        
        # Short circuit so that we can use autoreloader and debug
        edev_found = EndDeviceAdapter.fetch_by_property("sFDI", edev.sFDI)
        if edev_found:
            break
        
        EndDeviceAdapter.add(edev)
        
        # Add the end device to the list.
        index = EndDeviceAdapter.fetch_index(edev)
        
        
        EndDeviceAdapter.add_replace_child(edev, hrefs.END_DEVICE_REGISTRATION, 
                                   m.Registration(href=hrefs.registration_href(index), pIN=dev.pin, dateTimeRegistered=ts))
        edev.RegistrationLink = m.RegistrationLink(href=hrefs.registration_href(index))
        
        di = hrefs.EdevHref(edev_index=index, edev_subtype=hrefs.EDevSubType.DeviceInformation)
        EndDeviceAdapter.add_replace_child(edev, hrefs.END_DEVICE_INFORMATION, m.DeviceInformation(href=str(di)))
        edev.DeviceInformationLink = m.DeviceInformationLink(str(di))
        
        ds = hrefs.EdevHref(edev_index=index, edev_subtype=hrefs.EDevSubType.DeviceStatus)
        EndDeviceAdapter.add_replace_child(edev, hrefs.END_DEVICE_STATUS, m.DeviceStatus(str(ds)))
        edev.DeviceStatusLink = m.DeviceStatusLink(str(ds))
        
        #edev.FunctionSetAssignmentsListLink = m.FunctionSetAssignmentsListLink(href=hrefs.fsa_href(edev_index=index))
        
        fsa_programs = []
        for cfg_program in dev.programs:
            for program in programs:
                program.mRID = uuid_2030_5()
                if cfg_program["description"] == program.description:
                    fsa_programs.append(program)
                
        if len(fsa_programs) > 0:
            
            fsa = m.FunctionSetAssignments()
            
            FSAAdapter.add(fsa)
                            
            for derp in fsa_programs:                  
                FSAAdapter.add_replace_child(fsa, hrefs.FSA, derp)
                
            edev.FunctionSetAssignmentsListLink = m.FunctionSetAssignmentsListLink(href=hrefs.fsa_href(edev_index=index))
            
            # TODO we are hardcoding assuming only one fsa here.
            fsa.DERProgramListLink = m.DERProgramListLink(href=f"{hrefs.fsa_href(index=0)}_{hrefs.DER_PROGRAM}")
            # fsa = FSAAdapter.create(fsa_programs)
            # edev.FunctionSetAssignmentsListLink = m.FunctionSetAssignmentsListLink(href=hrefs.fsa_href(edev_index=index))
            # self._fsa.append(fsa)
            EndDeviceAdapter.add_replace_child(edev, hrefs.FSA, FSAAdapter)
            
        if dev.ders:
            der_href = hrefs.EdevHref(index, hrefs.EDevSubType.DER)
            deradapter = Adapter[m.DER](str(der_href), generic_type=m.DER)
            edev.DERListLink = m.DERListLink(str(der_href))
            
            EndDeviceAdapter.add_replace_child(edev, hrefs.DER, deradapter)
            for der_indx, der_cfg in enumerate(dev.ders):
                der_href = hrefs.EdevHref(edev_index=index, edev_subtype=hrefs.EDevSubType.DER, edev_subtype_index=der_indx)
                der = m.DER(href=str(der_href))
                der_href.edev_der_subtype = hrefs.DERSubType.Availability
                der.DERAvailabilityLink = m.DERAvailabilityLink(str(der_href))
                
                der_href.edev_der_subtype = hrefs.DERSubType.Capability
                der.DERCapabilityLink = m.DERCapabilityLink(str(der_href))
                
                der_href.edev_der_subtype = hrefs.DERSubType.Settings
                der.DERSettingsLink = m.DERSettingsLink(str(der_href))
                
                der_href.edev_der_subtype = hrefs.DERSubType.Status
                der.DERStatusLink = m.DERStatusLink(str(der_href))
                
                # Configure a link to the current program for the der.
                cfg_der_program = der_cfg.get("program")
                if cfg_der_program:
                    for derp_index, derp in enumerate(programs):
                        if cfg_der_program == derp.description:
                            der.CurrentDERProgramLink = m.CurrentDERProgramLink(derp.href)
                            break
                
                deradapter.add(der)
    
    ready_signal.send(EndDeviceAdapter)
        #self._end_devices.append(edev)
                        

        # log = m.LogEventList(href=hrefs.get_log_list_href(index),
        #                      all=0,
        #                      results=0,
        #                      pollRate=BaseAdapter.server_config().log_event_list_poll_rate)
        # edev.LogEventListLink = m.LogEventListLink(href=log.href)
        # add_href(log.href, log)

        # cfg = m.Configuration(href=hrefs.get_configuration_href(index))
        # add_href(cfg.href, cfg)
        # edev.ConfigurationLink = m.ConfigurationLink(cfg.href)

        # ps = m.PowerStatus(href=hrefs.get_power_status_href(index))
        # add_href(ps.href, ps)
        # edev.PowerStatusLink = m.PowerStatusLink(href=ps.href)

        # ds = m.DeviceStatus(href=hrefs.get_device_status(index))
        # add_href(ds.href, ds)
        # edev.DeviceStatusLink = m.DeviceStatusLink(href=ds.href)

        # di = m.DeviceInformation(href=hrefs.get_device_information(index))
        # add_href(di.href, di)
        # edev.DeviceInformationLink = m.DeviceInformationLink(href=di.href)

        # ts = int(round(datetime.utcnow().timestamp()))
        # reg = m.Registration(href=hrefs.get_registration_href(index),
        #                      pIN=dev.pin,
        #                      dateTimeRegistered=ts)
        # add_href(reg.href, reg)
        # edev.RegistrationLink = m.RegistrationLink(reg.href)

        # log = m.LogEventList(href=hrefs.get_log_list_href(index), all=0)
        # add_href(log.href, log)
        # edev.LogEventListLink = m.LogEventListLink(log.href)

        # fsa_list = m.FunctionSetAssignmentsList(href=hrefs.get_fsa_list_href(edev.href))

        # fsa = m.FunctionSetAssignments(href=hrefs.get_fsa_href(fsa_list_href=fsa_list.href,
        #                                                        index=0),
        #                                mRID="0F")
        # edev.FunctionSetAssignmentsListLink = m.FunctionSetAssignmentsListLink(fsa_list.href)

        # der_program_list = m.DERProgramList(href=hrefs.get_der_program_list(fsa_href=fsa.href),
        #                                     all=0,
        #                                     results=0)

        # fsa.DERProgramListLink = m.DERProgramListLink(href=der_program_list.href)
        # fsa_list.FunctionSetAssignments.append(fsa)

        # for cfg_program in dev.programs:
        #     for program in programs:
        #         program.mRID = "1F"
        #         if cfg_program["description"] == program.description:
        #             der_program_list.all += 1
        #             der_program_list.results += 1
        #             der_program_list.DERProgram.append(program)
        #             break

        # # Allow der list here
        # # # TODO: instantiate from config file.
        # der_list = m.DERList(
        #     href=hrefs.get_der_list_href(index),
        # #pollRate=900,
        #     results=0,
        #     all=0)
        # edev.DERListLink = m.DERListLink(der_list.href)

        # self._end_devices.append(edev)

        # edev_list.EndDevice.append(edev)
ready_signal.connect(initialize_end_device_adapter, DERProgramAdapter)




from __future__ import annotations

from threading import Lock
from typing import Optional, List
from uuid import uuid4

from ieee_2030_5.server.exceptions import AlreadyExistsError

__write_lock__: Lock = Lock()


class UUIDHandler:
    handler: UUIDHandler = None
    bag: dict = {}
    uuids: set = set()

    def __new__(cls):
        if UUIDHandler.handler is None:
            UUIDHandler.handler = super().__new__(cls)
        return UUIDHandler.handler

    def add_known(self, uuid: str, obj: object):
        assert isinstance(uuid, str) and obj is not None

        with __write_lock__:
            self.bag[uuid] = obj
            # TODO: Not sure this is correct for id(obj) to be able to determine equality.
            self.bag[id(obj)] = uuid
            self.uuids.add(uuid)

    def add(self, obj) -> str:
        """
        Add an object to the UUIDHandler.  If the object already exists
        in the collection then raise AlreadyExistsError

        :param: obj - The object to store in the handler.
        """
        if obj in self.bag:
            raise AlreadyExistsError(f"obj {obj} already exists in bag")

        new_uuid = self.generate()
        self.add_known(new_uuid, obj)
        return new_uuid

    def get_uuid(self, obj) -> Optional[str]:
        """
        Retrieve a uuid for a matching object.  If match exists the
        function returns the uuid, if not then returns None.

        :param: object An object to match.

        return: A string uuid or None
        """
        return self.bag.get(id(obj))

    def get_obj(self, uuid: str) -> Optional[object]:
        """
        Retrieve an object based on the passed uuid.  If match exists the
        function returns the object, if not then returns None.

        :param: str A uuid to match against.

        return: An object or None
        """
        return self.bag.get(uuid)

    def get_uuids(self) -> List[str]:
        return list(self.uuids.copy())

    def generate(self) -> str:
        new_uuid = str(uuid4())
        while new_uuid in self.uuids:
            new_uuid = str(uuid4())
        return new_uuid



import logging
from dataclasses import dataclass, field
from typing import Container, List, Optional, Sized, Tuple

import ieee_2030_5.hrefs as hrefs
import ieee_2030_5.models as m
from ieee_2030_5.adapters import BaseAdapter, ReturnCode
from ieee_2030_5.data.indexer import add_href, get_href

_log = logging.getLogger(__name__)

__all__: List[str] = [
    "MirrorUsagePointAdapter"
]

@dataclass
class _UsagePointWrapper:
    usage_point: m.UsagePoint   
    meter_readings: List[m.MeterReading] = field(default_factory=list)
    mirror_meter_readings: List[m.MirrorMeterReading] = field(default_factory=list)
    
    def fetch_reading_by_mRID(self, mRID) -> m.MeterReading:
        for x in self.meter_readings:
            if x.mRID == mRID:
                return x
        raise StopIteration()
        #return next(filter(lambda x: x.mRID == mRID, self.meter_readings))
    
    def fetch_mirror_meter_reading_index(self, mRID: str) -> int:
        for i, v in enumerate(self.mirror_meter_readings):
            if v.mRID == mRID:
                return i
        raise StopIteration()
        # return next(i for i, v in enumerate(self.mirror_meter_readings) if v.mRID == mRID)
    
    
@dataclass
class _UsagePointContainer(Container, Sized):
    __usage_points__: List[_UsagePointWrapper] = field(default_factory=list)
    
    def create_or_replace_reading(self, usage_point: m.UsagePoint, mirror_meter_reading: m.MirrorMeterReading) -> Tuple[ReturnCode, str]:
        
        assert usage_point in self, "Passed usage_point not contained within container"
        
        try:
            wrapper = self._fetch_wrapper_by_mRID(usage_point.mRID)
        except StopIteration:
            _log.error(f"Wrapper not found for usage point {usage_point}")
        else:
            try:
                mmr_index = wrapper.fetch_mirror_meter_reading_index(mirror_meter_reading.mRID)
                
                mirror_meter_reading.href = wrapper.mirror_meter_readings[mmr_index].href
                wrapper.mirror_meter_readings[mmr_index] = mirror_meter_reading                        
                return ReturnCode.NO_CONTENT, mirror_meter_reading.href
                
            except StopIteration:
                if not mirror_meter_reading.ReadingType:
                    _log.error(f"No ReadingType specified in meter reading.")
                    return ReturnCode.BAD_REQUEST.value, ""
                _log.debug(f"Reading not found for mRID {mirror_meter_reading.mRID}... creating")
                
                mr_href = hrefs.usage_point_href(usage_point.href, True)
                mr_reading_href = hrefs.usage_point_href(usage_point.href, True, len(wrapper.meter_readings))
                mr_type_href = hrefs.usage_point_href(usage_point_index=usage_point.href,
                                                      meter_reading_list = True, 
                                                      meter_reading_index=len(wrapper.meter_readings), 
                                                      meter_reading_type=True)
                meter_reading = m.MeterReading(href=mr_href,
                               ReadingTypeLink=m.ReadingTypeLink(href=mr_type_href),
                               ReadingLink=m.ReadingLink(mr_reading_href))
                mirror_meter_reading.href = mr_reading_href
                
                wrapper.meter_readings.append(meter_reading)                
                wrapper.mirror_meter_readings.append(mirror_meter_reading)
                
                return ReturnCode.CREATED.value, mr_reading_href
                
    
    def create_or_replace(self, mirror_usage_point: m.MirrorUsagePoint) -> m.UsagePoint:
        
        if mirror_usage_point in self:
            
            upt = self.fetch_by_mRID(mirror_usage_point.mRID)
            if not isinstance(upt, m.UsagePoint):
                raise ValueError("blah")
            upt.description=mirror_usage_point.description,
            upt.deviceLFDI=mirror_usage_point.deviceLFDI,
            upt.version=mirror_usage_point.version,
            upt.serviceCategoryKind=mirror_usage_point.serviceCategoryKind,
            upt.status=mirror_usage_point.status
            
        else:
            # 1. Since creating we can use the length of the __usage_points__ for determining the 
            #    next href
            upt = m.UsagePoint(href=hrefs.usage_point_href(len(UsagePointContainer)),
                            description=mirror_usage_point.description,
                            deviceLFDI=mirror_usage_point.deviceLFDI,
                            version=mirror_usage_point.version,
                            mRID=mirror_usage_point.mRID,
                            serviceCategoryKind=mirror_usage_point.serviceCategoryKind,
                            status=mirror_usage_point.status)
                        
            wrapper = _UsagePointWrapper(usage_point=upt)            
            self.__usage_points__.append(wrapper)
            
        if mirror_usage_point.MirrorMeterReading:
            wrapper = self._fetch_wrapper_by_mRID(mirror_usage_point.mRID)
            for reading in mirror_usage_point.MirrorMeterReading:
                self.create_or_replace_reading(upt, reading)
                
        return upt
        
    def fetch_list(self, start: int = 0, after: int = 0, limit: int = -1) -> m.UsagePointList:
        # if limit < 0:
        #     limit = len(self.__usage_points__)
        # elif limit > len(self.__usage_points__):
        #     limit = len(self.__usage_points__)
        
        # points = self.__usage_points__[]
        
        uptl = m.UsagePointList(href=hrefs.usage_point_href(),
                                UsagePoint=[upw.usage_point for upw in self.__usage_points__],
                                all=len(self.__usage_points__),
                                results=len(self.__usage_points__))
        return uptl
    
    def fetch_by_href(self, href: str) -> m.UsagePoint:
        for x in self.__usage_points__:
            if x.usage_point.href == href:
                return x.usage_point
        raise StopIteration()
        # return next(filter(lambda x: x.href == href, [y.usage_point for y in self.__usage_points__]))
    
    def fetch_by_mRID(self, mRID: str) -> m.UsagePoint:
        for x in self.__usage_points__:
            if x.usage_point.mRID == mRID:
                return x.usage_point
        
        raise StopIteration()
        # return next(filter(lambda x: x.mRID == mRID, [y.usage_point for y in self.__usage_points__]))
    
    def _fetch_wrapper_by_mRID(self, mRID: str) -> _UsagePointWrapper:
        for x in self.__usage_points__:
            if x.usage_point.mRID == mRID:
                return x
        
        raise StopIteration()
    
    def _fetch_wrapper_by_href(self, href: str) -> _UsagePointWrapper:
        for x in self.__usage_points__:
            if x.usage_point.href == href:
                return x
        
        raise StopIteration()
        #return next(filter(lambda x: x.usage_point.href == href, self.__usage_points__))

    def __contains__(self, other: object) -> bool:
        for x in self.__usage_points__:
            if x.usage_point.mRID == other.mRID:
                return True
        return False
        # return list(filter(lambda x: x.usage_point.mRID == other.mRID, self.__usage_points__))
    
    def __len__(self) -> int:
        return len(self.__usage_points__)

UsagePointContainer = _UsagePointContainer()

@dataclass
class MirrorUsagePointContainer(Container):
    __mirror_usage_points__: List[m.MirrorUsagePoint] = field(default_factory=list)
    
    def __contains__(self, __x: object) -> bool:
        return list(filter(lambda x: x.mRID == __x.mRID, self.__mirror_usage_points__))


@dataclass
class MeterReadingContainer(Container):
    __meter_readings__: List[m.MeterReading] = field(default_factory=list)
    
    def __contains__(self, __x: object) -> bool:
        return list(filter(lambda x: x.mRID == __x.mRID, self.__meter_readings__))
    
class _MirrorUsagePointAdapter:
    
    def __init__(self):
        self.__upt_container__: _UsagePointContainer = UsagePointContainer
        self.__mirror_usage_points__: List[m.MirrorUsagePoint] = []
    
    def fetch_usage_point_by_href(self, href: str) -> m.UsagePoint:
        return self.__upt_container__.fetch_by_href(href)
    
    def fetch_mirror_usage_point_list(self, start=0, after=0, limit=1) -> m.MirrorUsagePointList:
        return m.MirrorUsagePointList(href=hrefs.mirror_usage_point_href(), all=len(self.__mirror_usage_points__), results=len(self.__mirror_usage_points__),
                                      MirrorUsagePoint=self.__mirror_usage_points__, pollRate=BaseAdapter.server_config().usage_point_post_rate)
    
    def fetch_mirror_usage_by_href(self, href) -> m.MirrorUsagePoint:
        return next(filter(lambda x: x.href == href, self.__mirror_usage_points__))
    
    def get_list(self,start: Optional[int] = None,
                 after: Optional[int] = None,
                 length: Optional[int] = None) -> m.MirrorUsagePointList:
        if start is not None and after is not None:
            # after takes precedence
            index = after + 1 + start
        elif start is not None:
            index = start
        elif after is not None:
            index = after + 1
        else:
            index = 0
        value = get_href(href=hrefs.mirror_usage_point_href(index))

        if not value:
            mupl = m.MirrorUsagePointList(href=hrefs.mirror_usage_point_href())
            mupl.all = 0
            add_href(mupl.href, mupl)

        return get_href(href=hrefs.mirror_usage_point_href(index))

    
    def create(self, mup: m.MirrorUsagePoint) -> Tuple[ReturnCode, str]:
        """Creates a MirrorUsagePoint and its associated usage point
        
        This method creates all of the sub elements of the usage point and
        mirror usage points as well.
        """
        
        before = len(self.__upt_container__)
        upt = self.__upt_container__.create_or_replace(mup)
        after = len(self.__upt_container__)
        if after > before:
            # TODO: Don't hard code here.
            mup.href = upt.href.replace('upt', 'mup')
            mup.postRate = BaseAdapter.server_config().usage_point_post_rate
            self.__mirror_usage_points__.append(mup)
            return ReturnCode.CREATED.value, mup.href
        else:
            for i, o in enumerate(self.__mirror_usage_points__):
                if o.mRID == mup.mRID:
                    mup.href = o.href
                    self.__mirror_usage_points__[i] = mup
                    break
            return ReturnCode.NO_CONTENT.value, mup.href
        
    
    def __len__(self):
        return len(UsagePointContainer.__usage_points__)

    
    def create_reading(self, href: str, data: m.MirrorMeterReading) -> ReturnCode:
        """Create/replace reading passed to the method.
        
        Args:
            
            href: The pathinfo from the http request
            data: MirrorMeterReading to be added.
        """
        
        # 1. The href should be a specific mirror usage point or this should be forbidden.
        #    In addition, we use this as the base url for the upt url that is returned after
        #    the reading is created.
        try:
            #pths = href.split(hrefs.SEP)
            #mup = self.__mirror_usage_points__[pths[1]]
            href = href.replace(hrefs.MUP, hrefs.UTP)
            upt = self.__upt_container__.fetch_by_href(href)
            assert upt
            result = self.__upt_container__.create_or_replace_reading(upt, data)
        except StopIteration:
            # Not found within the list.
            return ReturnCode.BAD_REQUEST.value, ""
        
        return result
        
        # # 2. Attempt to find the mirror reading of the passed data.
        # try:
        #     rdng = next(filter(lambda x: x.mRID==data.mRID, MirrorUsagePointAdapter.__mirror_readings__))
        #     meter_reading: m.MirrorMeterReading = rdng.obj
        # except StopIteration:
        #     # If no ReadingType in the MirrorMeterReading then that is an error to create a new reading. 
        #     if not data.ReadingType:
        #         return ReturnCode.BAD_REQUEST
            
            
            
            
        
            
        
        # # 1. Find the usage point that this reading belongs to
        # mapped_usage_point = None
        # upls = UsagePointAdapter.fetch_all()        
        
        # for up in UsagePointAdapter.fetch_all().UsagePoint:
        #     if up.mRID == data.mRID:
        #         mapped_usage_point = up
        #         break
        
        # # If there is no usage point then the client should create one
        # if not mapped_usage_point:
        #     return ReturnCode.BAD_REQUEST
        
        # # 2. If the meter reading mRID matches an existing meter reading 
        # meter_readings_href = hrefs.SEP.join([mapped_usage_point.href, "mr"])
        # meter_readings = get_href(meter_readings_href)
        
        # if meter_readings is None:
        #     meter_readings = m.MeterReadingList(href=meter_readings_href, all=0)
        
        # next_meter_reading = meter_readings.all + 1
        
        
        # if isinstance(data, m.MeterReading):   
        #     data.href = hrefs.SEP.join([meter_readings_href, str(next_meter_reading)])
        #     meter_readings.MeterReading.append(data)
        # else:
        #     raise NotImplementedError
        
        
        
        # for index, rs in enumerate(data.MirrorReadingSet):
        #     rs_href = hrefs.SEP.join([mapped_usage_point.href, "mr", str(index)])
        #     for reading_index, reading in rs.Reading:
        #         reading_href = hrefs.SEP.join([rs_href], "rs", str(reading_index))
        #         reading.localID = reading_index
                
                
        # # TODO overwrite if reading set not found
        # mr_href = hrefs.SEP.join([mapped_usage_point.href, "mr"])
        # mr = get_href(mr_href)
        
        # if mr is None:
        #     mr = m.MeterReadingList(href=mr_href)                        
        #     mapped_usage_point.MeterReadingListLink = mr_href
        #     add_href(mr_href, mr)
        
        # next_index = len(mr.MeterReading) + 1
        
            
            
        

        
        # usage_point_href = hrefs.usage_point_href(uph.mirror_usage_point_index, meter_reading_list=True)
        
        # upl: m.UsagePointList = get_href(usage_point_href)
        
        
        # #upl.UsagePoint.append()        
        # # Look up the meter reading list
        # mr = get_href(href)
        
        
            
            
        

        # _log.debug(href)
        # _log.debug(data)
        
    
    
    def get_by_mRID(self,mRID: str) -> Optional[m.MirrorUsagePoint]:
        for mup in MirrorUsagePointAdapter.fetch_all().MirrorUsagePoint:
            if mup.mRID == mRID:
                return mup
        return None

    
    def get_by_index(self,index: int) -> m.MirrorUsagePoint:
        return get_href(hrefs.mirror_usage_point_href(index))

    
    def fetch_all(self) -> m.MirrorUsagePointList:
        mupl = get_href(hrefs.mirror_usage_point_href())

        if mupl is None:
            mupl = m.MirrorUsagePointList(href=hrefs.mirror_usage_point_href(),
                                          results=0,
                                          all=0,
                                          pollRate=30)
            add_href(mupl.href, mupl)

        return mupl
    
MirrorUsagePointAdapter = _MirrorUsagePointAdapter()



if __name__ == '__main__':
    
    mup = m.MirrorUsagePoint(mRID="foo",
                             roleFlags="09",
                             serviceCategoryKind=0,
                             status=1,
                             deviceLFDI="lfdididididiidididiididiididid")
    
    print("Creating mirror usage point")
    ret = MirrorUsagePointAdapter.create(mup)

    print("expected 201 and /mup_0")
    print(ret)
    
    mmr = m.MirrorMeterReading(mRID="29403bc10000000f0000d431",
                               description="Reactive Power",
                               lastUpdateTime="1677083028",
                               Reading=m.Reading(timePeriod=m.DateTimeInterval(start=0, duration=0)),
                               ReadingType=m.ReadingType(accumulationBehaviour=12,
                                                         commodity=1,
                                                         dataQualifier=0,
                                                         flowDirection=19,
                                                         kind=37,
                                                         powerOfTenMultiplier=0,
                                                         uom=63))
    
    print("Posting mirror reading to the wrong usage point should get 400, '' as response")
    # We know this is the first usage point created so post to the mirror usage point.
    ret2 = MirrorUsagePointAdapter.create_reading("/upt_0", mmr)
    print(ret2)
    ret3 = MirrorUsagePointAdapter.create_reading("/upt_0", mmr)
    print(ret3)
    
    
    ret3 = MirrorUsagePointAdapter.create(mup)
    
    print(ret2)


from typing import Optional

from flask import Response, request

import ieee_2030_5.adapters as adpt
import ieee_2030_5.hrefs as hrefs
import ieee_2030_5.models as m
from ieee_2030_5.server.base_request import RequestOp


class DERRequests(RequestOp):
    """
    Class supporting end devices and any of the subordinate calls to it.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get(self) -> Response:
        
        if not request.path.startswith(hrefs.DEFAULT_DER_ROOT):
            raise ValueError(f"Invalid path for {self.__class__} {request.path}")
        
        pth_split = request.path.split(hrefs.SEP)
        
        if len(pth_split) == 1:
            # TODO Add arguments
            value = adpt.DERAdapter.fetch_list()
        else:
            value = adpt.DERAdapter.fetch_at(int(pth_split[1]))

        return self.build_response_from_dataclass(value)
    

class DERProgramRequests(RequestOp):
    """
    Class supporting end devices and any of the subordinate calls to it.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get(self) -> Response:
        
        if not request.path.startswith(hrefs.DEFAULT_DERP_ROOT):
            raise ValueError("Invalid path passed to")
        
        derp_href = hrefs.DERProgramHref.parse(request.path)
        if derp_href.index == hrefs.NO_INDEX:
            retval = adpt.DERProgramAdapter.fetch_all(m.DERProgramList(href=request.path))
        
        else:
            derp = adpt.DERProgramAdapter.fetch(derp_href.index)
            
            if derp_href.derp_subtype == hrefs.DERProgramSubType.DERControlListLink:
                if derp_href.derp_subtype_index == hrefs.NO_INDEX:
                    retval = adpt.DERProgramAdapter.fetch_children(derp, hrefs.DERC, m.DERControlList(href=request.path))    
                else:
                    retval = adpt.DERProgramAdapter.fetch_child(derp, hrefs.DERC, derp_href.derp_subtype_index )    
            elif derp_href.derp_subtype == hrefs.DERProgramSubType.ActiveDERControlListLink:
                retval = adpt.DERProgramAdapter.fetch_children(derp, hrefs.DERCA, m.DERControlList(href=request.path))    
            # elif derp_href.der_subtype == hrefs.DERProgramSubType.DERCurveListLink:
            #     retval = adpt.DERProgramAdapter.fetch_der_ _active_control_list(derp_href.index)
            elif derp_href.derp_subtype == hrefs.DERProgramSubType.DefaultDERControlLink:
                retval = adpt.DERProgramAdapter.fetch_child(derp, hrefs.DDERC)
            else:
                raise ValueError(f"Found derph {derp_href}")
        
        return self.build_response_from_dataclass(retval)
    



from __future__ import annotations

import pickle
from copy import deepcopy
from dataclasses import dataclass, field
import logging

from datetime import datetime
from email.utils import format_datetime
from typing import Dict, Optional, List
from ieee_2030_5.persistance.points import set_point, get_point

__all__: List[str] = [
    "get_href",
    "add_href",
    "get_href_all_names",
    "get_href_filtered"
]

_log = logging.getLogger(__name__)


@dataclass
class Index:
    href: str
    item: object
    added: str  # Optional[Union[datetime | str]]
    last_written: str  # Optional[Union[datetime | str]]
    last_hash: Optional[int]


@dataclass
class Indexer:
    __items__: Dict = field(default=None)

    def init(self):
        if self.__items__ is None:
            self.__items__ = {}

    @property
    def length(self) -> int:
        self.init()
        return len(self.__items__)

    def add(self, href: str, item: dataclass):
        self.init()

        cached = self.__items__.get(href)
        if cached and cached.item == item:
            _log.debug(f"Item already cached {href}")
        else:
            added = format_datetime(datetime.utcnow())
            serialized_item = pickle.dumps(item)  # serialize_dataclass(item, serialization_type=SerializeType.JSON)
            obj = Index(href, item, added=added, last_written=added, last_hash=hash(serialized_item))
            # serialized_obj = serialize_dataclass(obj, serialization_type=SerializeType.JSON)

            # note storing Index object.
            set_point(href, pickle.dumps(obj))  # serialize_dataclass(obj, serialization_type=SerializeType.JSON))
            self.__items__[href] = obj

    def get(self, href) -> dataclass:
        self.init()
        if href in self.__items__:
            index = pickle.loads(get_point(href))  # pickle.loads(get_point(href))
            # index = pickle.loads(self.__items__.get(href))
            # index = deserialize_dataclass(data, SerializeType.JSON)
            data = index.item
        else:
            data = None

        return data

    def get_all(self) -> List:
        return deepcopy([x.item for x in self.__items__.values()])


__indexer__ = Indexer()


def add_href(href: str, item: dataclass):
    __indexer__.add(href, item)


def get_href(href: str) -> dataclass:
    return __indexer__.get(href)


def get_href_filtered(href_prefix: str) -> List[dataclass] | []:
    if __indexer__.__items__ is None:
        return []
    
    return [v.item for k, v in __indexer__.__items__.items()
            if k.startswith(href_prefix) and v.item is not None]


def get_href_all_names():
    return [x for x in __indexer__.__items__.keys()
            if __indexer__.__items__[x] is not None]



import logging
from typing import Optional

import werkzeug.exceptions
from flask import Response, request

import ieee_2030_5.hrefs as hrefs
import ieee_2030_5.models as m
from ieee_2030_5.adapters import Adapter
from ieee_2030_5.adapters.enddevices import EndDeviceAdapter
from ieee_2030_5.adapters.fsa import FSAAdapter
from ieee_2030_5.data.indexer import get_href
from ieee_2030_5.models import Registration
from ieee_2030_5.server.base_request import RequestOp
from ieee_2030_5.types_ import Lfdi
from ieee_2030_5.utils import dataclass_to_xml, xml_to_dataclass

_log = logging.getLogger(__name__)

class EDevRequests(RequestOp):
    """
    Class supporting end devices and any of the subordinate calls to it.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        
    def put(self) -> Response:
        parsed = hrefs.EdevHref.parse(request.path)
        
        mysubobj = xml_to_dataclass(request.data.decode('utf-8'))
        
        
        ed = EndDeviceAdapter.fetch(parsed.edev_index)
        deradapter: Adapter[m.DER] = EndDeviceAdapter.fetch_child(ed, hrefs.DER)
        der = deradapter.fetch(parsed.edev_subtype_index)
        
        
        try:
            deradapter.fetch_child(der, parsed.edev_der_subtype.value)
            response_status = 204
        except (KeyError, IndexError):
            response_status = 201
        finally:
            if not mysubobj.href:
                mysubobj.href = request.path
            deradapter.add_replace_child(der, parsed.edev_der_subtype.value, mysubobj)
            
        return Response(status=response_status)
        # response_code = adpt.DERAdapter.store(parsed, xml_to_dataclass(request.data.decode('utf-8')))
        #return Response(status=int(response_code))                                              
                                              
                        
        
    def post(self, path: Optional[str] = None) -> Response:
        """
        Handle post request to /edev
        
        The expectation is that data will be an xml object like the following:
        
            <EndDevice xmlns="urn:ieee:std:2030.5:ns">
                <sFDI>231589308001</sFDI>
                <changedTime>0</changedTime>
            </EndDevice>
        
        Args:
            path: 

        Returns:

        """
        # request.data should have xml data.
        if not request.data:
            raise werkzeug.exceptions.Forbidden()

        ed: m.EndDevice = xml_to_dataclass(request.data.decode('utf-8'))

        if not isinstance(ed, m.EndDevice):
            raise werkzeug.exceptions.Forbidden()

        # This is what we should be using to get the device id of the registered end device.
        device_id = self.tls_repo.find_device_id_from_sfdi(ed.sFDI)
        ed.lFDI = self.tls_repo.lfdi(device_id)
        if end_device := EndDeviceAdapter.fetch_by_lfdi(ed.lfdi):
            status = 200
            ed_href = end_device.href
        else:
            if not ed.href:
                ed = EndDeviceAdapter.store(device_id, ed)

            ed_href = ed.href
            status = 201

        return Response(status=status, headers={'Location': ed_href})

    def get(self) -> Response:
        """
        Supports the get request for end_devices(EDev) and end_device_list_link.

        Paths:
            /edev
            /edev/0
            /edev/0/di
            /edev/0/rg
            /edev/0/der

        """
        _log.debug(f"Args are: {request.args}")
        edev_href = hrefs.EdevHref.parse(request.path)
        start = int(request.args.get("s", 0))
        limit = int(request.args.get("l", 1))
        after = int(request.args.get("a", 0))
        

        ed = EndDeviceAdapter.fetch_by_property('lFDI', self.lfdi)
        
        # means we don't have any /edev without any index
        if edev_href.edev_subtype is hrefs.EDevSubType.None_Available:
            retval = m.EndDeviceList(href=request.path, all=1, results=1, EndDevice=[ed])
        
        
        elif edev_href.edev_subtype is hrefs.EDevSubType.FunctionSetAssignments:
            
            fsaadpt: FSAAdapter = EndDeviceAdapter.fetch_child(ed, hrefs.FSA, edev_href.edev_index)
            
            if edev_href.edev_subtype_index == hrefs.NO_INDEX:
                retval = fsaadpt.fetch_all(m.FunctionSetAssignmentsList(href=request.path), start=start, after=after, limit=limit)
            else:
                retval = fsaadpt.fetch(edev_href.edev_subtype_index)
                
            
            # if edev_href.edev_subtype_index == hrefs.NO_INDEX:
            #     retval = EndDeviceAdapter.fetch_children(m.DERList(request.path), hrefs.DERSubType)
            # else:
            #     retval = EndDeviceAdapter.fetch_child(ed, hrefs.FSA, edev_href.edev_subtype_index)
                
        # we have /edev_index_der or /edev_index_der_subindex or /edev_index_der_subindex_dersubtype
        elif edev_href.edev_subtype == hrefs.EDevSubType.DER:
        
            deradpt: Adapter[m.DER] = EndDeviceAdapter.fetch_child(ed, hrefs.DER, edev_href.edev_index)
                        
            if edev_href.edev_subtype_index == hrefs.NO_INDEX:
                retval = deradpt.fetch_all(m.DERList(href=request.path), start=start, after=after, limit=limit)
            else:
                der = deradpt.fetch(edev_href.edev_subtype_index)
                if edev_href.edev_der_subtype is hrefs.DERSubType.None_Available:
                    retval = der
                else:
                    retval = deradpt.fetch_child(der, edev_href.edev_der_subtype.value)
        
            # try:
            #     ed = EndDeviceAdapter.fetch(int(pth_split[1]))
            #     # FSA is a list off the end device, the rest are singleton items.
            #     if pth_split[2] == hrefs.FSA:                   
            #         retval = EndDeviceAdapter.fetch_children(ed, hrefs.FSA, m.FunctionSetAssignmentsList(href=request.path))
            #     elif pth_split[2] == hrefs.DER:
            #         deradpt = EndDeviceAdapter.fetch_child(ed, hrefs.DER, )
            #         retval = EndDeviceAdapter.fetch_children(ed, hrefs.DER, m.DERList(href=request.path))
            #     else:
            #         retval = EndDeviceAdapter.fetch_child(ed, pth_split[2], 0)
            # except KeyError:
            #     raise werkzeug.exceptions.NotFound("Missing Resource")
            
        else:
            retval = EndDeviceAdapter.fetch_child(ed, edev_href.edev_subtype.value)
            # if pth_split[2] == "rg":
            #     retval = EndDeviceAdapter.fetch_registration(edev_index=int(pth_split[1]))
            # elif pth_split[2] == "di":
            #     retval = "foo"
            # elif pth_split[2] == "fsa":
            #     retval = EndDeviceAdapter.fetch_fsa_list(edev_index=int(pth_split[1]))
            # elif pth_split[2] == "der":
            #     retval = adpt.DERAdapter.fetch_list(edev_index=int(pth_split[1]))
            
        return self.build_response_from_dataclass(retval)


class SDevRequests(RequestOp):
    """
    SelfDevice is an alias for the end device of a client.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get(self) -> Response:
        """
        Supports the get request for end_devices(EDev) and end_device_list_link.

        Paths:
            /sdev

        """
        end_device = self._end_devices.get_end_device_list(self.lfdi).EndDevice[0]
        return self.build_response_from_dataclass(end_device)

class FSARequests(RequestOp):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        
    def get(self):
        """ Retrieve a FSA or Program List
        """
        
        fsa_href = hrefs.fsa_parse(request.path)
        
        if fsa_href.fsa_index == hrefs.NO_INDEX:
            retval = FSAAdapter.fetch_all(m.FunctionSetAssignmentsList(), "FunctionSetAssignments")
        elif fsa_href.fsa_sub == hrefs.FSASubType.DERProgram.value:
            fsa = FSAAdapter.fetch(fsa_href.fsa_index)
            retval = FSAAdapter.fetch_children(fsa, "fsa", m.DERProgramList())
            # retval = FSAAdapter.fetch_children_list_container(fsa_href.fsa_index, m.DERProgram, m.DERProgramList(href="/derp"), "DERProgram")
        else:
            retval = FSAAdapter.fetch(fsa_href.fsa_index)
            
        # pth_split = request.path.split(hrefs.SEP)
        
        
        # if len(pth_split) == 1:
        #     retval = FSAAdapter.fetch_list()
        # elif len(pth_split) == 2:
        #     retval = FSAAdapter.fetch_at(int(pth_split[1]))
        # elif len(pth_split) == 3:
        #     retval = EndDeviceAdapter.fetch_fsa_list(edev_index=int(pth_split[1]))
        # elif len(pth_split) == 4:
        #     retval = EndDeviceAdapter.fetch_fsa(edev_index=int(pth_split[1]), fsa_index=int(pth_split[3]))
        # else:
        #     raise ValueError(f"Path split is {pth_split}")
            
        return self.build_response_from_dataclass(retval)


import logging
from functools import lru_cache
from typing import Dict, List

import ieee_2030_5.hrefs as hrefs
import ieee_2030_5.models as m
from ieee_2030_5.adapters import Adapter, AdapterListProtocol, BaseAdapter
from ieee_2030_5.types_ import Lfdi

_log = logging.getLogger(__name__)


__all__: List[str] = [
    "FSAAdapter"
]


FSAAdapter = Adapter[m.FunctionSetAssignments](url_prefix=hrefs.fsa_href(), generic_type=m.FunctionSetAssignments)


import enum
""" Metering
"""


class RtgNormalCategoryType(enum.IntEnum):
    not_specified = 0
    category_a = 1
    category_b = 2


class RtgAbnormalCategoryType(enum.IntEnum):
    not_specified = 0
    category_I = 1
    category_II = 2
    category_III = 3


class DataQualifierType(enum.IntEnum):
    Not_applicable = 0
    Average = 2
    Maximum = 8
    Minimum = 9
    Normal = 12
    Standard_deviation_of_population = 29
    Standard_deviation_of_sample = 30


class CommodityType(enum.IntEnum):
    Not_applicable = 0
    Electricity_secondary_metered = 1
    Electricity_primary_metered = 2
    Air = 4
    NaturalGas = 7
    Propane = 8
    PotableWater = 9
    Steam = 10
    WasteWater = 11
    HeatingFluid = 12
    CoolingFluid = 13


class FlowDirectionType(enum.IntEnum):
    Not_applicable = 0
    Forward = 1
    Reverse = 19


class UomType(enum.IntEnum):
    Not_applicable = 0
    Amperes = 5
    Kelvin = 6
    Degrees_celsius = 23
    Voltage = 29
    Joule = 31
    Hz = 33
    W = 38
    M_cubed = 42
    VA = 61
    VAr = 63
    CosTheta = 65
    V_squared = 67
    A_squared = 69
    VAh = 71
    Wh = 72
    VArh = 73
    Ah = 106
    Ft_cubed = 119
    Ft_cubed_per_hour = 122
    M_cubed_per_hour = 125
    US_gallons = 128
    UG_gallons_per_hour = 129
    Imperial_gallons = 130
    Imperial_gallons_per_hour = 131
    BTU = 132
    BTU_per_hour = 133
    Liter = 134
    Liter_per_hour = 137
    PA_gauge = 140
    PA_absolute = 155
    Therm = 169


class RoleFlagsType(enum.Flag):
    IsMirror = 0
    IsPremiseAggregationPoint = 1
    IsPEV = 2
    IsDER = 4
    IsRevenueQuality = 8
    IsDC = 16
    IsSubmeter = 32


class AccumlationBehaviourType(enum.IntEnum):
    Not_applicable = 0
    Cumulative = 3
    DeltaData = 4
    Indicating = 6
    Summation = 9
    Instantaneous = 12


class ServiceKind(enum.IntEnum):
    Electricity = 0
    Gas = 1
    Water = 2
    Time = 3
    Pressure = 4
    Heat = 5
    Cooling = 6


class QualityFlagsType(enum.Flag):
    Valid = 0
    Manually_edited = 1
    estimated_using_reference_day = 2
    estimated_using_linear_interpolation = 4
    questionable = 8
    derived = 16
    projected = 32


# p163
class ConsumptionBlockType(enum.IntEnum):
    Not_applicable = 0
    Block_1 = 1
    Block_2 = 2
    Block_3 = 3
    Block_4 = 4
    Block_5 = 5
    Block_6 = 6
    Block_7 = 7
    Block_8 = 8
    Block_9 = 9
    Block_10 = 10
    Block_11 = 11
    Block_12 = 12
    Block_13 = 13
    Block_14 = 14
    Block_15 = 15
    Block_16 = 16


# p170
class TOUType(enum.IntEnum):
    Not_applicable = 0
    TOU_A = 1
    TOU_B = 2
    TOU_C = 3
    TOU_D = 4
    TOU_E = 5
    TOU_F = 6
    TOU_G = 7
    TOU_H = 8
    TOU_I = 9
    TOU_J = 10
    TOU_K = 11
    TOU_L = 12
    TOU_M = 13
    TOU_N = 14
    TOU_O = 15


class KindType(enum.IntEnum):
    Not_applicable = 0
    Currency = 3
    Demand = 8
    Energy = 12
    Power = 37


class PhaseCode(enum.IntEnum):
    Not_applicable = 0
    Phase_C = 32    # and S2
    Phase_CN = 33    # and S2N
    Phase_CA = 40
    Phase_B = 64
    Phase_BN = 65
    Phase_BC = 66
    Phase_A = 128    # and S1
    Phase_AN = 129    # and S1N
    Phase_AB = 132
    Phase_ABC = 224


""" Subscription/Notification
"""


class ResponseRequiredType(enum.Flag):
    enddevice_shall_indicate_that_message_was_received = 0
    enddevice_shall_indicate_specific_response = 1
    enduser_customer_response_is_required = 2


class SubscribableType(enum.IntEnum):
    resource_does_not_support_subscriptions = 0
    resource_supports_non_conditional_subscriptions = 1
    resource_supports_conditional_subscriptions = 2
    resource_supports_both_conditional_and_non_conditional_subscriptions = 3



import inspect
import logging
import typing
from dataclasses import dataclass, fields, is_dataclass
from enum import Enum
from typing import (Any, Dict, Generic, List, Optional, Protocol, Type,
                    TypeVar, get_args, get_origin)

from blinker import Signal

import ieee_2030_5.config as cfg
import ieee_2030_5.hrefs as hrefs
import ieee_2030_5.models as m
from ieee_2030_5.certs import TLSRepository
from ieee_2030_5.models.sep import List_type

_log = logging.getLogger(__name__)

class AlreadyExists(Exception):
    pass

class ReturnCode(Enum):
    OK = 200
    CREATED = 201
    NO_CONTENT = 204
    BAD_REQUEST = 400


def populate_from_kwargs(obj: object, **kwargs) -> Dict[str, Any]:

    if not is_dataclass(obj):
        raise ValueError(f"The passed object {obj} is not a dataclass.")

    for k in fields(obj):
        if k.name in kwargs:
            type_eval = eval(k.type)

            if typing.get_args(type_eval) is typing.get_args(Optional[int]):
                setattr(obj, k.name, int(kwargs[k.name]))
            elif typing.get_args(k.type) is typing.get_args(Optional[bool]):
                setattr(obj, k.name, bool(kwargs[k.name]))
            # elif bytes in args:
            #     setattr(obj, k.name, bytes(kwargs[k.name]))
            else:
                setattr(obj, k.name, kwargs[k.name])
            kwargs.pop(k.name)
    return kwargs

class AdapterIndexProtocol(Protocol):
    
    def fetch_at(self, index: int) -> m.Resource:
        pass


class AdapterListProtocol(AdapterIndexProtocol):
    def fetch_list(self, start: int = 0, after: int =0, limit:int = 0) -> m.List_type:
        pass
    
    def fetch_edev_all(self) -> List:
        pass
        

ready_signal = Signal("ready-signal")

T = TypeVar('T')
C = TypeVar('C')
D = TypeVar('D')


class Adapter(Generic[T]):
    
    def __init__(self, url_prefix: str, **kwargs):
        if "generic_type" not in kwargs:
            raise ValueError("Missing generic_type parameter")
        self._generic_type: Type = kwargs['generic_type']
        self._href_prefix: str = url_prefix
        self._current_index: int  = -1
        self._item_list: Dict[int, T] = {}
        self._child_prefix: Dict[Type, str] = {}
        self._child_map: Dict[int, Dict[str, List[C]]] = {}
        
    @property
    def href_prefix(self) -> str:
        return self._href_prefix
    
    @property
    def href(self) -> str:
        return self._href_prefix
    
    @href.setter
    def href(self, value:str) -> None:
        self._href_prefix = value
    
    def fetch_by_property(self, prop: str, prop_value: Any) -> Optional[T]:
        for obj in self._item_list.values():
            if getattr(obj, prop) == prop_value:
                return obj
            
    def fetch_child_names(self) -> List[str]:
        return list(self._child_prefix.keys())
    
    def add_container(self, child_type: Type, href_prefix: str):
        self._child_prefix[child_type] = href_prefix
        
    def remove_child(self, parent: T, name: str, child: Any):
        found_index = self.fetch_index(parent)
        self._child_map[found_index][name].remove(child)
        
    def remove_child_by_mrid(self, parent: T, name: str, mRID: str):
        
        found_index = self.fetch_index(parent)
        
        indexes = [index for index, x in enumerate(self._child_map[found_index][name]) if x.mRID == mRID]
        for index in sorted(indexes, reverse=True):
            self._child_map[found_index][name].pop(index)
        
    def add_replace_child(self, parent: T, name: str, child: Any, href: str = None):
        
        # Make sure parent is in the Adapter by looking for it's index.
        found_index = self.fetch_index(parent)
        
        # Deal with missing indexes
        if found_index not in self._child_map:
            self._child_map[found_index] = {}
            
        if name not in self._child_map[found_index]:
            self._child_map[found_index][name] = []
            
        if len(self._child_map[found_index][name]) > 0:
            if not isinstance(child, type(self._child_map[found_index][name][0])):
                raise ValueError(f"Children can only have single types {type(child)} != {type(self._child_map[found_index][name][0])}")
        if not child.href:
            if href:
                child.href = href
            else:
                child.href = hrefs.SEP.join([parent.href, name, str(len(self._child_map[found_index][name]))])
        
        # Replace based upon resource href
        for index, c in enumerate(self._child_map[found_index][name]):
            if c.href == child.href:
                _log.debug(f"Replacing child {child.href}")
                self._child_map[found_index][name][index] = child
                return
            
        self._child_map[found_index][name].append(child)
        
    def fetch_children_by_parent_index(self, parent_index: int, child_type: Type) -> List[Type]:
        if child_type not in self._child_map[parent_index]:
            raise KeyError(f"No child object of type {child_type}")
        
        return self._child_map[parent_index][child_type]
    
    def fetch_children(self, parent: T, name: str, container: Optional[Type] = None) -> List[Type]:
        found_index = self.fetch_index(parent)
        # Should end with List if container is not None
        if container is not None:
            if not container.__class__.__name__.endswith("List"):
                raise ValueError(f"Invalid container, type must end in List")
        
        try:
            children = self._child_map[found_index][name]
        except KeyError:
            children = []
        
        retval = children
        
        if container is not None:
            prop = container.__class__.__name__[:container.__class__.__name__.find("List")]
            setattr(container, prop, children)
            setattr(container, "results", len(children))
            setattr(container, "all", len(children))
            retval = container
            
        return retval
                
    def fetch_child(self, parent: T, name: str, index: int = 0) -> Type:
        return self.fetch_children(parent, name)[index]
            
    def add(self, item: T):
        if not isinstance(item, self._generic_type):
            raise ValueError(f"Item {item} is not of type {self._generic_type}")
        
        # Only replace if href is specified.
        if hasattr(item, 'href') and getattr(item, 'href') is None:
            setattr(item, 'href', hrefs.SEP.join([self._href_prefix, str(self._current_index + 1)]))
        self._current_index += 1
        self._item_list[self._current_index] = item
    
    def fetch_all(self, container: Optional[D] = None, start: int = 0, after: int = 0, limit: int = 1) -> D:

        if container is not None:
            if not container.__class__.__name__.endswith("List"):
                raise ValueError("Must have List as the last portion of the name for instance")
        
            prop_found = container.__class__.__name__[:container.__class__.__name__.find("List")]
        
            items = list(self._item_list.values())
            all_len = len(items)
            all_results = len(items)
            all_items = items
                
            if start > len(items):
                all_items = []
                all_results = 0
            else:
                if limit == 0:
                    all_items = items[start:]
                else:
                    all_items = items[start: start + limit]
                all_results = len(all_items)
                
            setattr(container, prop_found, all_items)
            setattr(container, "all", all_len)
            setattr(container, "results", all_results)
        else:
            container = list(self._item_list.values())
            
        return container
    
    def fetch_index(self, obj: T, using_prop: str = None) -> int:
        found_index = -1
        for index, obj1 in self._item_list.items():
            if using_prop is None:    
                if obj1 == obj:
                    found_index = index
                    break
            else:
                if getattr(obj, using_prop) == getattr(obj1, using_prop):
                    found_index = index
                    break
        if found_index == -1:
            raise KeyError(f"Object {obj} not found in adapter")
        return found_index
    
    def fetch(self, index: int):
        return self._item_list[index]
    
    def fetch_by_mrid(self, mRID: str):
        for item in self._item_list:
            if not hasattr(item, 'mRID'):
                raise ValueError(f"Item of {type(T)} does not have mRID property")
            if item.mRID == mRID:
                return item
            
        raise KeyError()
    
    def size(self) -> int:
        return len(self._item_list)
    
    def size_children(self, parent: T, name: str) -> int:
        return len(self.fetch_children(parent, name))
    
    def fetch_child_index_by_mrid(self, parent: T, name: str, mRID: str) -> int:
        for index, child in enumerate(self.fetch_children(parent, name)):
            if child.mRID == mRID:
                return index
        
        raise KeyError("mRID not found")
    
    def replace_child(self, parent: T, name: str, index: int, child: Any):
        children = self.fetch_children(parent, name)
        if not type(child) == type(children[index]):
            return ValueError(f"Children should be  of the same type {type(child)} is not {type(children[index])}")
        parent_index = self.fetch_index(parent)
        if not child.href:
            child.href = children[index].href
        self._child_map[parent_index][name][index] = child
    
    
    
            

class BaseAdapter:
    __count__: int = 0
    __server_configuration__: cfg.ServerConfiguration
    __tls_repository__: cfg.TLSRepository = None
    __lfdi__mapped_configuration__: Dict[str, cfg.DeviceConfiguration] = {}
    after_initialized = Signal('after-initialized')
        
    @classmethod
    def get_next_index(cls) -> int:
        """Retrieve the next index for an adapter list."""
        return cls.__count__

    @classmethod
    def increment_index(cls) -> int:
        """Increment the list to the next index and return the result to the caller.
        
        
        """
        next = cls.get_next_index()
        cls.__count__ += 1
        return next
    
    @classmethod
    def ready(cls) -> Signal:
        return Signal("ready")

    @classmethod
    def get_current_index(cls) -> int:
        return cls.__count__

    @staticmethod
    def server_config() -> cfg.ServerConfiguration:
        return BaseAdapter.__server_configuration__

    @staticmethod
    def device_configs() -> List[cfg.DeviceConfiguration]:
        return BaseAdapter.__server_configuration__.devices

    @staticmethod
    def tls_repo() -> cfg.TLSRepository:
        return BaseAdapter.__tls_repository__

    @staticmethod
    def get_config_from_lfdi(lfdi: str) -> Optional[cfg.DeviceConfiguration]:
        return BaseAdapter.__lfdi__mapped_configuration__.get(lfdi)

    @staticmethod
    def is_initialized():
        return BaseAdapter.__device_configurations__ is not None and BaseAdapter.__tls_repository__ is not None

    @staticmethod
    def initialize(server_config: cfg.ServerConfiguration, tlsrepo: TLSRepository):
        """Initialize all of the adapters
        
        The initialization means that there are concrete object backing the storage system based upon
        urls that can be read during the http call to the spacific end point.  In other words a
        DERCurve dataclass can be retrieved from storage by going to the href /dc/1 rather than
        having to get it through an object.  
        
        The adapters are responsible for storing data into the object store using add_href function.
        """
        BaseAdapter.__server_configuration__ = server_config
        BaseAdapter.__lfdi__mapped_configuration__ = {}
        BaseAdapter.__tls_repository__ = tlsrepo

        # Map from the configuration id and lfdi to the device configuration.
        for cfg in server_config.devices:
            lfdi = tlsrepo.lfdi(cfg.id)
            BaseAdapter.__lfdi__mapped_configuration__[lfdi] = cfg

        #BaseAdapter.after_initialized.send(BaseAdapter)
        ready_signal.send(BaseAdapter)
        # BaseAdapter.ready().send(BaseAdapter)
        # Find subclasses of us and initialize them calling _initalize method
        # TODO make this non static
        #EndDeviceAdapter._initialize()

    @staticmethod
    def build(**kwargs) -> dataclass:
        raise NotImplementedError()

    @staticmethod
    def store(value: dataclass) -> dataclass:
        raise NotImplementedError()

    @staticmethod
    def build_instance(cls, cfg_dict: Dict, signature_cls=None) -> object:
        if signature_cls is None:
            signature_cls = cls
        return cls(**{
            k: v
            for k, v in cfg_dict.items() if k in inspect.signature(signature_cls).parameters
        })

from ieee_2030_5.adapters.dcap import DeviceCapabilityAdapter
from ieee_2030_5.adapters.der import (DERControlAdapter, DERCurveAdapter,
                                      DERProgramAdapter)
from ieee_2030_5.adapters.enddevices import EndDeviceAdapter
from ieee_2030_5.adapters.log import LogAdapter
from ieee_2030_5.adapters.mupupt import MirrorUsagePointAdapter



import time
from datetime import datetime
from threading import Thread

from blinker import Signal


class _TimeAdapter(Thread):
    tick = Signal("tick")
            
    @staticmethod
    def user_readable(timestamp: int) -> str:
        dt = datetime.fromtimestamp(timestamp)
        return dt.isoformat() # .strftime("%m/%d/%Y, %H:%M:%S")
    
    @staticmethod
    def from_iso(iso_fmt_date: str) -> int:
        dt = datetime.strptime(iso_fmt_date, "%Y-%m-%dT%H:%M:%S")
        return int(time.mktime(dt.timetuple()))
    
    def run(self) -> None:
        
        while True:
            self._tick = int(time.mktime(datetime.utcnow().timetuple()))
            _TimeAdapter.tick.send(self._tick)
            time.sleep(1)
        
    
TimeAdapter = _TimeAdapter()
TimeAdapter.daemon = True
TimeAdapter.start()



# -------------------------------------------------------------------------------
# Copyright (c) 2022, Battelle Memorial Institute All rights reserved.
# Battelle Memorial Institute (hereinafter Battelle) hereby grants permission to any person or entity
# lawfully obtaining a copy of this software and associated documentation files (hereinafter the
# Software) to redistribute and use the Software in source and binary forms, with or without modification.
# Such person or entity may use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and may permit others to do so, subject to the following conditions:
# Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimers.
# Redistributions in binary form must reproduce the above copyright notice, this list of conditions and
# the following disclaimer in the documentation and/or other materials provided with the distribution.
# Other than as used herein, neither the name Battelle Memorial Institute or Battelle may be used in any
# form whatsoever without the express written consent of Battelle.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
# BATTELLE OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
# General disclaimer for use with OSS licenses
#
# This material was prepared as an account of work sponsored by an agency of the United States Government.
# Neither the United States Government nor the United States Department of Energy, nor Battelle, nor any
# of their employees, nor any jurisdiction or organization that has cooperated in the development of these
# materials, makes any warranty, express or implied, or assumes any legal liability or responsibility for
# the accuracy, completeness, or usefulness or any information, apparatus, product, software, or process
# disclosed, or represents that its use would not infringe privately owned rights.
#
# Reference herein to any specific commercial product, process, or service by trade name, trademark, manufacturer,
# or otherwise does not necessarily constitute or imply its endorsement, recommendation, or favoring by the United
# States Government or any agency thereof, or Battelle Memorial Institute. The views and opinions of authors expressed
# herein do not necessarily state or reflect those of the United States Government or any agency thereof.
#
# PACIFIC NORTHWEST NATIONAL LABORATORY operated by BATTELLE for the
# UNITED STATES DEPARTMENT OF ENERGY under Contract DE-AC05-76RL01830
# -------------------------------------------------------------------------------

import logging
import os
import socket
import sys
import threading
from argparse import ArgumentParser
from multiprocessing import Process
from pathlib import Path
from time import sleep

import yaml
from werkzeug.serving import BaseWSGIServer

import ieee_2030_5.hrefs as hrefs
from ieee_2030_5.certs import TLSRepository
from ieee_2030_5.config import InvalidConfigFile, ServerConfiguration
from ieee_2030_5.data.indexer import add_href
from ieee_2030_5.flask_server import run_server
from ieee_2030_5.server.server_constructs import initialize_2030_5

_log = logging.getLogger()


class ServerThread(threading.Thread):

    def __init__(self, server: BaseWSGIServer):
        threading.Thread.__init__(self, daemon=True)
        self.server = server

    def run(self):
        _log.info(f'starting server on {self.server.host}:{self.server.port}')
        self.server.serve_forever()

    def shutdown(self):
        _log.info("shutting down server")
        self.server.shutdown()


def get_tls_repository(cfg: ServerConfiguration,
                       create_certificates_for_devices: bool = True) -> TLSRepository:
    tlsrepo = TLSRepository(cfg.tls_repository,
                            cfg.openssl_cnf,
                            cfg.server_hostname,
                            cfg.proxy_hostname,
                            clear=create_certificates_for_devices,
                            generate_admin_cert=cfg.generate_admin_cert)

    if create_certificates_for_devices:
        already_represented = set()

        # registers the devices, but doesn't initialize_device the end devices here.
        for k in cfg.devices:
            if k in already_represented:
                _log.error(f"Already have {k.id} represented by {k.device_category_type}")
            else:
                already_represented.add(k)
                tlsrepo.create_cert(k.id)
                _log.debug(
                    f"for {k.id}\nlfdi -> {tlsrepo.lfdi(k.id)}\nsfdi -> {tlsrepo.sfdi(k.id)}")
    return tlsrepo


def _shutdown():
    make_stop_file()
    sleep(1)
    remove_stop_file()


def should_stop():
    return Path('server.stop').exists()


def make_stop_file():
    with open('server.stop', 'w') as w:
        pass


def remove_stop_file():
    pth = Path('server.stop')
    if pth.exists():
        os.remove(pth)


def _run_ui():
    
    os.system('python gui/spa/main.py')
    

def _main():
    parser = ArgumentParser()

    parser.add_argument(dest="config", help="Configuration file for the server.")
    parser.add_argument("--no-validate",
                        action="store_true",
                        help="Allows faster startup since the resolving of addresses is not done!")
    parser.add_argument(
        "--no-create-certs",
        action="store_true",
        help="If specified certificates for for client and server will not be created.")
    parser.add_argument("--debug", action="store_true", help="Debug level of the server")
    parser.add_argument("--production",
                        action="store_true",
                        default=False,
                        help="Run the server in a threaded environment.")
    opts = parser.parse_args()

    logging_level = logging.DEBUG if opts.debug else logging.INFO
    logging.basicConfig(level=logging_level)

    os.environ["IEEE_2030_5_CONFIG_FILE"] = str(
        Path(opts.config).expanduser().resolve(strict=True))

    cfg_dict = yaml.safe_load(Path(opts.config).expanduser().resolve(strict=True).read_text())

    config = ServerConfiguration(**cfg_dict)

    if config.lfdi_mode == "lfdi_mode_from_file":
        os.environ["IEEE_2030_5_CERT_FROM_COMBINED_FILE"] = '1'

    assert config.tls_repository
    assert len(config.devices) > 0
    assert config.server_hostname

    add_href(hrefs.get_server_config_href(), config)
    unknown = []
    # Only check for resolvability if not passed --no-validate
    if not opts.no_validate:
        _log.debug("Validating hostnames and/or ip of devices are resolvable.")
        for i in range(len(config.devices)):
            assert config.devices[i].hostname

            try:
                socket.gethostbyname(config.devices[i].hostname)
            except socket.gaierror:
                if hasattr(config.devices[i], "ip"):
                    try:
                        socket.gethostbyname(config.devices[i].ip)
                    except socket.gaierror:
                        unknown.append(config.devices[i].hostname)
                else:
                    unknown.append(config.devices[i].hostname)

    if unknown:
        _log.error("Couldn't resolve the following hostnames.")
        for host in unknown:
            _log.error(host)
        sys.exit(1)

    create_certs = not opts.no_create_certs
    tls_repo = get_tls_repository(config, create_certs)

    # Initialize the repository of 2030.5 devices.
    end_devices = initialize_2030_5(config, tls_repo)

    #if not opts.production:
    try:
        # p = Process(target = _run_ui)
        # p.daemon = True
        # p.start()
                
        run_server(config,
                    tls_repo,
                    end_devices,
                    debug=opts.debug,
                    use_reloader=True,
                    use_debugger=opts.debug,
                    threaded=False)
    except KeyboardInterrupt:
        _log.info("Shutting down server")
    # else:
    #     server = build_server(config, tls_repo, enddevices=end_devices)

    #     thread = None
    #     try:
    #         remove_stop_file()
    #         thread = ServerThread(server)
    #         thread.start()
    #         while not should_stop():
    #             sleep(0.5)
    #     except KeyboardInterrupt as ex:
    #         _log.info("Exiting program.")
    #     finally:
    #         if thread:
    #             thread.shutdown()
    #             thread.join()


if __name__ == '__main__':
    try:
        _main()
    except InvalidConfigFile as ex:
        print(ex.args[0])
    except KeyboardInterrupt:
        pass



import logging
from typing import List

import ieee_2030_5.models as m
from ieee_2030_5.adapters import BaseAdapter
from ieee_2030_5.data.indexer import add_href, get_href

__all__: List[str] = [
    "LogAdapter"
]

_log = logging.getLogger(__name__)
class _LogAdapter:
    
    def __init__(self):
        pass        
    
    def __after_base_init__(self, sender):        
        print("after base init")
        print(sender)

    @staticmethod
    def store(path: str, logevent: m.LogEvent):
        """Store a logevent to the given path.
        
        The 2030.5 Logevent is based upon a specific device so /edev/edevid/log is the event
        list that should be stored.  The store method only stores a single event at a time.  The
        2030.5 standard says we should hold at least 10 logs per logevent level."""
        event_list: m.LogEventList = get_href(path)
        if event_list is None:
            event_list = m.LogEventList(
                href=path, pollRate=BaseAdapter.server_config().log_event_list_poll_rate)
            event_list.LogEvent = []

        event_list.LogEvent.append(logevent)
        event_list.LogEvent = sorted(event_list.LogEvent, key="createdDateTime")
        add_href(path, event_list)

    @staticmethod
    def fetch_list(path: str, start: int = 0, after: int = 0, limit: int = 1) -> m.LogEventList:
        # TODO: implement start length
        event_list: m.LogEventList = get_href(path)

        if event_list is None:
            return m.LogEventList(href=path,
                                  all=0,
                                  results=0,
                                  pollRate=BaseAdapter.server_config().log_event_list_poll_rate)

        event_list.all = len(event_list.LogEvent)
        event_list.results = len(event_list.LogEvent)

        return event_list

LogAdapter = _LogAdapter()
BaseAdapter.after_initialized.connect(LogAdapter.__after_base_init__)


import argparse
import hashlib
import logging
import os
import sys
from pathlib import Path
from typing import List, Optional, Tuple

from cryptography import x509
from cryptography.hazmat.backends import default_backend

__all__ = ['TLSRepository']

from ieee_2030_5.types_ import Lfdi, PathStr
from ieee_2030_5.utils.tls_wrapper import OpensslWrapper, TLSWrap

_log = logging.getLogger(__name__)


def lfdi_from_fingerprint(fingerprint: str) -> Lfdi:
    fp = fingerprint.replace(":", "")
    return Lfdi(fp[:40])


def sfdi_from_lfdi(lfdi: Lfdi) -> int:
    assert len(lfdi) == 40, "lfdi must be 160-bits (40 hex characters) long."
    hex_str = str(int(lfdi[:9], 16))
    check_bit = 0
    full_sum = sum([int(x) for x in hex_str])
    while not (full_sum + check_bit) % 10 == 0:
        check_bit += 1
    return int(hex_str + str(check_bit))


class TLSRepository:

    def __init__(self,
                 repo_dir: PathStr,
                 openssl_cnffile_template: PathStr,
                 serverhost: str,
                 proxyhost: str = None,
                 clear=False,
                 **kwargs):
        if isinstance(repo_dir, str):
            repo_dir = Path(repo_dir).expanduser().resolve()
        if isinstance(openssl_cnffile_template, str):
            openssl_cnffile_template = Path(openssl_cnffile_template).expanduser().resolve()
            if not openssl_cnffile_template.exists():
                raise ValueError(f"openssl_cnffile does not exist {openssl_cnffile_template}")
        self._repo_dir = repo_dir
        self._certs_dir = repo_dir.joinpath("certs")
        self._private_dir = repo_dir.joinpath("private")
        self._combined_dir = repo_dir.joinpath("combined")
        self._openssl_cnf_file = self._repo_dir.joinpath(openssl_cnffile_template.name)
        self._common_names = {serverhost: serverhost}
        if proxyhost:
            self._common_names[proxyhost] = proxyhost
        self._client_common_name_set = set()

        if not self._repo_dir.exists() or not self._certs_dir.exists() or \
                not self._private_dir.exists() or not self._combined_dir.exists():
            self._certs_dir.mkdir(parents=True)
            self._private_dir.mkdir(parents=True)
            self._combined_dir.mkdir(parents=True)

        index_txt = self._repo_dir.joinpath("index.txt")
        serial = self._repo_dir.joinpath("serial")
        if clear:
            for x in self._private_dir.iterdir():
                x.unlink()
            for x in self._certs_dir.iterdir():
                x.unlink()
            for x in self._combined_dir.iterdir():
                x.unlink()
            for x in self._repo_dir.iterdir():
                if x.is_file():
                    x.unlink()
            try:
                index_txt.unlink()
            except FileNotFoundError:
                pass
            try:
                serial.unlink()
            except FileNotFoundError:
                pass

        if not index_txt.exists():
            index_txt.write_text("")
        if not serial.exists():
            serial.write_text("01")

        new_contents = openssl_cnffile_template.read_text().replace(
            "dir             = /home/gridappsd/tls", f"dir = {repo_dir}")
        self._openssl_cnf_file.write_text(new_contents)
        self._ca_key = self._private_dir.joinpath("ca.pem")
        self._ca_cert = self._certs_dir.joinpath("ca.pem")
        self._serverhost = serverhost
        self._proxyhost = proxyhost

        self._tls: TLSWrap = OpensslWrapper(self._openssl_cnf_file)
        self._cert_paths: List[Path] = []

        # Create a new ca key if not exists.
        if not Path(self._ca_key).exists():
            self.__create_ca__()
            self._tls.tls_create_private_key(self.__get_key_file__(self._serverhost))
            self.create_cert(self._serverhost, True)
            if proxyhost:
                self.create_cert(self._proxyhost, True)

        if not clear:
            #_log.debug(list(self._certs_dir.glob('*.pem')))
            
            # creating certs has something screwy so we are going
            # to create the cert_paths based upon the private key
            # files.
            for f in self._private_dir.glob('*.pem'):
                f = Path(f)
                f = self._certs_dir.joinpath(f.name)
                cn = self.get_common_name(f.stem)
                if cn not in ('ca', serverhost):
                    self._client_common_name_set.add(cn)

        generate_admin_cert = kwargs.pop('generate_admin_cert', False)

        if generate_admin_cert:
            admin_key = self._private_dir.joinpath("admin.pem")
            admin_cert = self._certs_dir.joinpath("admin.pem")
            if not admin_key.exists():
                self._tls.tls_create_private_key(admin_key)
                self.create_cert(admin_cert.stem)

        for d in self._cert_paths:
            lfdi_from_stem = self.lfdi(d.stem)
            from_stem = self.sfdi(d.stem)
            _log.debug(f"For cert {d.name}\n\tLFDI: {lfdi_from_stem}\n\tSFDI: {from_stem}")

        if len(kwargs) > 0:
            raise ValueError(f"Not all kwargs used: {kwargs.keys()}")

    def __create_ca__(self):
        self._tls.tls_create_private_key(self._ca_key)
        self._tls.tls_create_ca_certificate("ca", self._ca_key, self._ca_cert)
        self._tls.tls_create_pkcs23_pem_and_cert(self._ca_key, self._ca_cert,
                                                 self.__get_combined_file__("ca"))

    def create_cert(self, common_name: str, as_server: bool = False):

        if not self.__get_key_file__(common_name).exists():
            self._tls.tls_create_private_key(self.__get_key_file__(common_name))

        self._tls.tls_create_signed_certificate(common_name, self._ca_key, self._ca_cert,
                                                self.__get_key_file__(common_name),
                                                self.__get_cert_file__(common_name), as_server)

        self._tls.tls_create_pkcs23_pem_and_cert(self.__get_key_file__(common_name),
                                                 self.__get_cert_file__(common_name),
                                                 self.__get_combined_file__(common_name))

        self._common_names[common_name] = common_name
        self._cert_paths.append(self.__get_cert_file__(common_name=common_name))

    def lfdi(self, device_id: str) -> Lfdi:
        """
        Using the fingerprint of the certifcate return the left truncation of 160 bits with no check digit.
        Example:
          From:
            3E4F-45AB-31ED-FE5B-67E3-43E5-E456-2E31-984E-23E5-349E-2AD7-4567-2ED1-45EE-213A
          Return:
            3E4F-45AB-31ED-FE5B-67E3-43E5-E456-2E31-984E-23E5
            as an integer.
        """
        # 160 / 4 == 40
        fp = self.fingerprint(device_id, True)
        return Lfdi(lfdi_from_fingerprint(fp))

    def sfdi(self, device_id: str) -> int:
        lfdi_ = self.lfdi(device_id)
        return sfdi_from_lfdi(lfdi_)

    def fingerprint(self, device_id: str, without_colan: bool = True) -> str:
        if os.environ.get('IEEE_2030_5_CERT_FROM_COMBINED_FILE'):
            # _log.debug("Using hash from combined file.")
            value = Path(self.__get_combined_file__(device_id)).read_text()
            value = hashlib.sha256(value.encode('utf-8')).hexdigest()
        else:
            value = self._tls.tls_get_fingerprint_from_cert(self.__get_cert_file__(device_id))
        if without_colan:
            value = value.replace(":", "")
        if "=" in value:
            value = value.split("=")[1]
        assert isinstance(value, str)
        return value

    def get_common_name(self, device_id: str) -> x509:
        pem_data = Path(self.__get_cert_file__(device_id)).read_bytes()
        cert = x509.load_pem_x509_certificate(pem_data, default_backend())
        return cert.subject.get_attributes_for_oid(x509.oid.NameOID.COMMON_NAME)[0].value

    def get_file_pair(self, device_id: str) -> Tuple[str, str]:
        """ Get cert, key from the repository based on passed device_id"""
        return (self.__get_cert_file__(device_id).as_posix(),
                self.__get_key_file__(device_id).as_posix())

    @property
    def client_list(self) -> List[str]:
        return list(self._client_common_name_set)

    @property
    def ca_key_file(self) -> Path:
        return self._ca_key

    @property
    def ca_cert_file(self) -> Path:
        return self._ca_cert

    @property
    def proxy_key_file(self) -> Path:
        if not self._proxyhost:
            raise ValueError("No proxy host available")
        return self.__get_key_file__(self._proxyhost)

    @property
    def proxy_cert_file(self) -> Path:
        if not self._proxyhost:
            raise ValueError("No proxy host available")
        return self.__get_cert_file__(self._proxyhost)

    @property
    def server_key_file(self) -> Path:
        return self.__get_key_file__(self._serverhost)

    @property
    def server_cert_file(self) -> Path:
        return self.__get_cert_file__(self._serverhost)

    def find_device_id_from_sfdi(self, sfdi: int) -> Optional[str]:
        """
        Searches the certificate paths for a device id that maps to the sfdi passed into the method.
        Args:
            sfdi:

        Returns:

        """
        device_id = None
        _log.debug(f"Attempting to find sfid: {sfdi}")
        for d in self._certs_dir.glob("*.pem"):
            try:
                if sfdi == self.sfdi(d.stem):
                    device_id = d.stem
                    break
            except FileNotFoundError:
                pass
        return device_id

    def __get_cert_file__(self, common_name: str) -> Path:
        return self._certs_dir.joinpath(f"{common_name}.pem")

    def __get_key_file__(self, common_name: str) -> Path:
        return self._private_dir.joinpath(f"{common_name}.pem")

    def __get_combined_file__(self, common_name: str) -> Path:
        return self._combined_dir.joinpath(f"{common_name}-combined.pem")


def _main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--dir", help="Directory of certificates determine lfdi and sfdi from.")
    parser.add_argument("--file", help="File to determine lfdi and sfdi from.")

    opts = parser.parse_args()

    if opts.dir and opts.file:
        sys.stderr.write("Only specify dir or file.\n")
        sys.exit(1)
    elif opts.dir is None and opts.file is None:
        sys.stderr.write("Must specify either --dir or --file.\n")
        sys.exit(1)

    target = opts.file
    if opts.dir:
        target = opts.dir

    target = Path(target)
    if not target.exists():
        sys.stderr.write("Invalid file or directory refrence.\n")
        sys.exit(1)

    if target.is_dir():
        targets = [target.glob("*.crt")]
    else:
        targets = [target]

    for t in targets:
        fingerprint = OpensslWrapper.tls_get_fingerprint_from_cert(t)
        lfdi = lfdi_from_fingerprint(fingerprint)
        sfdi = sfdi_from_lfdi(lfdi)
        sys.stdout.write(f"certificate: {t}\n")
        sys.stdout.write(f"-" * 60 + "\n")
        sys.stdout.write(f"fingerprint: {fingerprint}\n")
        sys.stdout.write(f"lfdi: {lfdi.decode('ascii')}\n")
        sys.stdout.write(f"sfdi: {sfdi}\n\n")


if __name__ == '__main__':

    logging.basicConfig(level=logging.DEBUG)

    fingerprint = "3E4F-45AB-31ED-FE5B-67E3-43E5-E456-2E31-984E-23E5-349E-2AD7-4567-2ED1-45EE-213A".replace(
        "-", "")
    lfdi = lfdi_from_fingerprint(fingerprint)
    sfdi = sfdi_from_lfdi(lfdi)
    print(f"fingerprint: {fingerprint}")
    print(f"lfdi: {lfdi}")
    print(f"sfdi: {sfdi}")

    fingerprint = "B5:65:B2:C4:D4:22:59:72:58:6E:4E:E2:B1:F2:98:D4:20:62:15:DB:53:49:AB:45:2F:D2:8F:BC:62:2C:28:1D".replace(
        ":", "")
    lfdi = lfdi_from_fingerprint(fingerprint)
    sfdi = sfdi_from_lfdi(lfdi)
    print(f"fingerprint: {fingerprint}")
    print(f"lfdi: {lfdi}")
    print(f"sfdi: {sfdi}")

    #
    # tlsrepo = TLSRepository(repo_dir="~/tls",
    #                         openssl_cnffile_template="../openssl.cnf",
    #                         clear=False,
    #                         serverhost="gridappsd_dev_2004:8443")
    # fingerprint = tlsrepo.fingerprint("dev1")
    # # fingerprint = "3F4F-45AB-31ED-FE5B-67E3-43E5-E456-2E31-984E-23E5-349E-2AD7-4567-2ED1-45EE-213B".replace("-", "")
    # print(len(fingerprint))
    # print("my lfdi: ", fingerprint[:40])
    # tlsrepo.sfdi_from_lfdi(fingerprint[:40].encode("ascii"))
    # # Each char is 4 bits so 9*4 == 36
    # print("left 36 bits: ", fingerprint[:9])
    # print("to int from fingerprint", int(fingerprint[:9], 16))
    # interum = str(int(fingerprint[:9], 16))
    # print(int(interum[-2:]))
    # add_value = 1
    # while not (int(interum[-2:]) + add_value) % 10 == 0:
    #     add_value += 1
    #
    # print(add_value)
    # interum = interum + str(add_value)
    # print(f"sfdi = ", interum)
    #
    # print(f" our sfdi: {tlsrepo.sfdi('dev1')}")
    #
    # _log.debug(f"fingerprint: {tlsrepo.fingerprint('dev1', False)}")
    #
    # _log.debug(f"dev1 lfdi: {tlsrepo.lfdi('dev1')}, sfdi: {tlsrepo.sfdi('dev1')}")
